library entity_store;

import 'dart:math' as math;

import 'package:fixnum/fixnum.dart';
import 'package:quiver/core.dart' as qcore;

import 'dart:collection' show SplayTreeMap;
import '../../src/schema_v1_pb2.dart';

class EntityStore {
  EntityGroup root = new EntityGroup(null);

  final Map<Int64, Key> keyIndex = new Map<Int64, Key>();

  void insert(Entity entity) {
    var path = root;
    for (var pathElement in entity.key.pathElement) {
      _checkSpecifiedPath(pathElement);
      if (path.children[pathElement] == null) {
        path.children[pathElement] = new EntityGroup(path);
      }
      path = path.children[pathElement];
    }
    path.entity = entity.clone();
  }

  EntityResult lookup(Key key) {
    var group = root;
    for (var pathElement in key.pathElement) {
      _checkSpecifiedPath(pathElement);
      if (group.children[pathElement] == null) {
        var missing = new Entity()
          ..key = key.clone();
        return new EntityResult()..entity = missing;
      }
      group = group.children[pathElement];
    }
    if (group.entity == null) {
      var missing = new Entity()
        ..key = key.clone();
      return new EntityResult()..entity = missing;
    }
    var found = group.entity.clone();
    return new EntityResult()..entity = found;
  }

  /**
   * Verify that the given path element has both a kind
   * and
   */
  _checkSpecifiedPath(Key_PathElement element) {
    if (!element.hasKind())
      throw 'Key_PathElement with no kind';
    if (!element.hasId() && !element.hasName())
      throw 'Not a completely specified path';
  }
}

class _Bucket {
  Set<Key> ids;

  _Bucket():
    this.ids = new Set<Key>();
}

class PropertyIdentifier {
  final String kind;
  final String name;

  PropertyIdentifier(this.kind, this.name);

  bool operator ==(Object other) =>
      other is PropertyIdentifier &&
      other.kind == kind &&
      other.name == name;

  int get hashCode => qcore.hash2(kind, name);
}

class Index<T> {
  final PropertyIdentifier prop;
  final SplayTreeMap<T,_Bucket> buckets;

  Index._(this.prop):
    this.buckets = new SplayTreeMap<T,_Bucket>();

  int compareValues(Value a, Value b) {
    if (a.hasBlobKeyValue() || b.hasBlobKeyValue())
      throw new UnsupportedError("Blob key value");
    if (a.hasEntityValue() || b.hasEntityValue())
      throw new UnsupportedError("Entity value");
    if (a.hasIntegerValue()
        || a.hasTimestampMicrosecondsValue())
      return _compareIntValue(a.integerValue, b);
    if (a.hasBooleanValue())
      return _compareBooleanValue(a.booleanValue, b);
    if (a.hasStringValue())
      return _compareStringValue(a.stringValue, b);
    if (a.hasDoubleValue())
      return _compareDoubleValue(a.doubleValue, b);
    if (a.hasKeyValue())
      return _compareKeyValue(a.keyValue, b);
    if (a.hasBlobValue())
      return _compareBlobValue(a.blobValue, b);
    return _compareNullValue(b);
  }

  int _compareNullValue(Value b) {
    if (b.hasIntegerValue()
        || b.hasBooleanValue()
        || b.hasStringValue()
        || b.hasDoubleValue()
        || b.hasKeyValue()
        || b.hasBlobValue())
      return 1;
    //b is null.
    return 0;
  }

  int _compareIntValue(Int64 value, Value b) {
    if (b.hasIntegerValue())
      return value.compareTo(b.integerValue);
    if (b.hasTimestampMicrosecondsValue())
      return value.compareTo(b.integerValue);
    if (b.hasStringValue()
        || b.hasDoubleValue()
        || b.hasStringValue()
        || b.hasKeyValue()
        || b.hasBlobValue())
      return -1;
    //b is null;
    return 1;
  }

  int _compareBooleanValue(bool value, Value b) {
    if (b.hasIntegerValue()
        || b.hasTimestampMicrosecondsValue())
      return 1;
    if (b.hasBooleanValue()) {
      var bValue = b.booleanValue;
      if (value) {
        return bValue ? 0 : 1;
      } else {
        return bValue ? -1: 0;
      }
    }
    if (b.hasStringValue() ||
        b.hasDoubleValue() ||
        b.hasKeyValue() ||
        b.hasBlobValue())
      return -1;
    //b is null.
    return 1;
  }

  int _compareStringValue(String value, Value b) {
    if (b.hasIntegerValue()
        || b.hasBooleanValue()) {
      return 1;
    }
    if (b.hasStringValue())
      return value.compareTo(b.stringValue);
    if (b.hasDoubleValue() ||
        b.hasKeyValue() ||
        b.hasBlobValue())
      return -1;
    //b is null.
    return 1;
  }

  int _compareDoubleValue(double value, Value b) {
    if (b.hasIntegerValue() ||
        b.hasBooleanValue() ||
        b.hasStringValue())
      return 1;
    if (b.hasDoubleValue())
      return value.compareTo(b.doubleValue);
    if (b.hasKeyValue() ||
        b.hasBlobValue())
      return -1;
    //b is null
    return 1;
  }

  int _compareKeyValue(Key value, Value b) {
    if (b.hasIntegerValue() ||
        b.hasBooleanValue() ||
        b.hasStringValue() ||
        b.hasDoubleValue())
      return 1;
    if (b.hasKeyValue()) {
      var bValue = b.keyValue();
      return _compareKeys(value, bValue);
    }
    if (b.hasBlobValue())
      return -1;
    //b is null
    return 1;
  }

  int _compareBlobValue(List<int> blobValue, Value b) {
    if (b.hasIntegerValue() ||
        b.hasBooleanValue() ||
        b.hasStringValue() ||
        b.hasDoubleValue() ||
        b.hasKeyValue())
      return 1;
    if (b.hasBlobValue()) {
      List<int> bValue = b.blobValue;
      var minLen = math.min(blobValue.length, bValue.length);
      for (var i=0;i<minLen;i++) {
        if (blobValue[i] < bValue[i])
          return -1;
        if (blobValue[i] > bValue[i])
          return 1;
      }
      if (minLen < blobValue.length)
        return -1;
      if (minLen < bValue.length)
        return 1;
      return 0;
    }
    //b is null
    return 1;
  }







}

class IntegerIndex extends Index<int> {
  IntegerIndex(PropertyIdentifier prop): super._(prop);
}

class DoubleIndex extends Index<double> {
  DoubleIndex(PropertyIdentifier prop): super._(prop);
}

class StringIndex extends Index<String> {
  StringIndex(PropertyIdentifier prop): super._(prop);
}


class EntityGroup {
  final EntityGroup parent;
  Entity entity;

  final Map<Key_PathElement, EntityGroup> children = new Map<Key_PathElement, EntityGroup>();

  final Map<PropertyIdentifier, Index> indexes = new Map<PropertyIdentifier,Index>();

  EntityGroup(this.parent);

  void updateIndexesForEntity(Entity ent) {
    var kind = ent.key.pathElement.last.kind;
    for (var prop in ent.property) {
      var propIdent = new PropertyIdentifier(kind, prop.name);
      if (prop.value.indexed) {
        if (!indexes.containsKey(propIdent)) {
          indexes[propIdent] = new Index(propIdent);
        }
        indexes[propIdent].add(prop);
      }
    }
  }

  void updateIndexes() {
    for (var child in children) {
      if (child.entity != null) {
        var ent = child.entity;
        var kind = ent.key.pathElement.last.kind;
        for (var prop in ent.entity.properties) {
          if (prop.value.indexed) {
            var propIdentifier = new PropertyIdentifier(kind, prop.name);
            if (indexes[propIdentifier] == null) {
              indexes[propIdentifier] = new Index(propIdentifier);
            }
          }
        }
      }
    }
  }
}